@@ -63,8 +63,6 @@ typedef enum
 /** For each warning type, its severity.  */
 static severity warnings_flag[warnings_size];
 
-static unsigned *indent_ptr = NULL;
-
 styled_ostream_t errstream = NULL;
 
 void
@@ -109,46 +107,59 @@ flush (FILE *out)
 | --warnings's handling.  |
 `------------------------*/
 
-static const char * const warnings_args[] =
+ARGMATCH_DEFINE_GROUP (warning, warnings)
+
+static const argmatch_warning_doc argmatch_warning_docs[] =
 {
-  "none",
-  "midrule-values",
-  "yacc",
-  "conflicts-sr",
-  "conflicts-rr",
-  "deprecated",
-  "empty-rule",
-  "precedence",
-  "other",
-  "all",
-  "error",
-  "everything",
-  0
+  { "conflicts-sr",     N_("S/R conflicts (enabled by default)") },
+  { "conflicts-rr",     N_("R/R conflicts (enabled by default)") },
+  { "deprecated",       N_("obsolete constructs") },
+  { "empty-rule",       N_("empty rules without %empty") },
+  { "midrule-values",   N_("unset or unused midrule values") },
+  { "precedence",       N_("useless precedence and associativity") },
+  { "yacc",             N_("incompatibilities with POSIX Yacc") },
+  { "other",            N_("all other warnings (enabled by default)") },
+  { "all",              N_("all the warnings except 'yacc'") },
+  { "no-CATEGORY",      N_("turn off warnings in CATEGORY") },
+  { "none",             N_("turn off all the warnings") },
+  { "error[=CATEGORY]", N_("treat warnings as errors") },
+  { NULL, NULL }
 };
 
-static const warnings warnings_types[] =
+static const argmatch_warning_arg argmatch_warning_args[] =
 {
-  Wnone,
-  Wmidrule_values,
-  Wyacc,
-  Wconflicts_sr,
-  Wconflicts_rr,
-  Wdeprecated,
-  Wempty_rule,
-  Wprecedence,
-  Wother,
-  Wall,
-  Werror,
-  Weverything
+  { "none",           Wnone },
+  { "midrule-values", Wmidrule_values },
+  { "yacc",           Wyacc },
+  { "conflicts-sr",   Wconflicts_sr },
+  { "conflicts-rr",   Wconflicts_rr },
+  { "deprecated",     Wdeprecated },
+  { "empty-rule",     Wempty_rule },
+  { "precedence",     Wprecedence },
+  { "other",          Wother },
+  { "all",            Wall },
+  { "everything",     Weverything },
+  { NULL, Wnone }
 };
 
-ARGMATCH_VERIFY (warnings_args, warnings_types);
+const argmatch_warning_group_type argmatch_warning_group =
+{
+  argmatch_warning_args,
+  argmatch_warning_docs,
+  N_("Warning categories include:"),
+  NULL
+};
+
+void
+warning_usage (FILE *out)
+{
+  argmatch_warning_usage (out);
+}
 
 void
 warning_argmatch (char const *arg, size_t no, size_t err)
 {
-  int value = XARGMATCH ("--warning", arg + no + err,
-                         warnings_args, warnings_types);
+  int value = *argmatch_warning_value ("--warning", arg + no + err);
 
   /* -Wnone == -Wno-everything, and -Wno-none == -Weverything.  */
   if (!value)
@@ -187,7 +198,14 @@ warning_argmatch (char const *arg, size_t no, size_t err)
 void
 warnings_argmatch (char *args)
 {
-  if (args)
+  if (!args)
+    warning_argmatch ("all", 0, 0);
+  else if (STREQ (args, "help"))
+    {
+      warning_usage (stdout);
+      exit (EXIT_SUCCESS);
+    }
+  else
     for (args = strtok (args, ","); args; args = strtok (NULL, ","))
       if (STREQ (args, "error"))
         warnings_are_errors = true;
@@ -203,8 +221,6 @@ warnings_argmatch (char *args)
 
           warning_argmatch (args, no, err);
         }
-  else
-    warning_argmatch ("all", 0, 0);
 }
 
 static const char*
@@ -292,6 +308,8 @@ complain_init_color (void)
 void
 complain_init (void)
 {
+  caret_init ();
+
   warnings warnings_default =
     Wconflicts_sr | Wconflicts_rr | Wdeprecated | Wother;
 
@@ -365,16 +383,17 @@ warning_is_enabled (warnings flags)
 static void
 warnings_print_categories (warnings warn_flags, FILE *out)
 {
-  for (size_t i = 0; warnings_args[i]; ++i)
-    if (warn_flags & warnings_types[i])
+  for (int wbit = 0; wbit < warnings_size; ++wbit)
+    if (warn_flags & (1 << wbit))
       {
-        severity s = warning_severity (warnings_types[i]);
+        warnings w = 1 << wbit;
+        severity s = warning_severity (w);
         const char* style = severity_style (s);
         fputs (" [", out);
         begin_use_class (style, out);
         fprintf (out, "-W%s%s",
                  s == severity_error ? "error=" : "",
-                 warnings_args[i]);
+                 argmatch_warning_argument (&w));
         end_use_class (style, out);
         fputc (']', out);
         /* Display only the first match, the second is "-Wall".  */
@@ -386,6 +405,7 @@ warnings_print_categories (warnings warn_flags, FILE *out)
  *
  * \param loc     the location, defaulting to the current file,
  *                or the program name.
+ * \param indent  optional indentation for the error message.
  * \param flags   the category for this message.
  * \param sever   to decide the prefix to put before the message
  *                (e.g., "warning").
@@ -397,26 +417,25 @@ warnings_print_categories (warnings warn_flags, FILE *out)
  */
 static
 void
-error_message (const location *loc, warnings flags, severity sever,
-               const char *message, va_list args)
+error_message (const location *loc, int *indent, warnings flags,
+               severity sever, const char *message, va_list args)
 {
-  unsigned pos = 0;
+  int pos = 0;
 
   if (loc)
     pos += location_print (*loc, stderr);
   else
-    pos += fprintf (stderr, "%s", current_file ? current_file : program_name);
+    pos += fprintf (stderr, "%s", grammar_file ? grammar_file : program_name);
   pos += fprintf (stderr, ": ");
 
-  if (indent_ptr)
+  if (indent)
     {
-      if (*indent_ptr)
+      if (*indent)
         sever = severity_disabled;
-      if (!*indent_ptr)
-        *indent_ptr = pos;
-      else if (*indent_ptr > pos)
-        fprintf (stderr, "%*s", *indent_ptr - pos, "");
-      indent_ptr = NULL;
+      if (!*indent)
+        *indent = pos;
+      else if (*indent > pos)
+        fprintf (stderr, "%*s", *indent - pos, "");
     }
 
   const char* style = severity_style (sever);
@@ -451,8 +470,8 @@ error_message (const location *loc, warnings flags, severity sever,
 /** Raise a complaint (fatal error, error or just warning).  */
 
 static void
-complains (const location *loc, warnings flags, const char *message,
-           va_list args)
+complains (const location *loc, int *indent, warnings flags,
+           const char *message, va_list args)
 {
   severity s = warning_severity (flags);
   if ((flags & complaint) && complaint_status < status_complaint)
@@ -462,7 +481,7 @@ complains (const location *loc, warnings flags, const char *message,
     {
       if (severity_error <= s && ! complaint_status)
         complaint_status = status_warning_as_error;
-      error_message (loc, flags, s, message, args);
+      error_message (loc, indent, flags, s, message, args);
     }
 
   if (flags & fatal)
@@ -474,23 +493,22 @@ complain (location const *loc, warnings flags, const char *message, ...)
 {
   va_list args;
   va_start (args, message);
-  complains (loc, flags, message, args);
+  complains (loc, NULL, flags, message, args);
   va_end (args);
 }
 
 void
-complain_indent (location const *loc, warnings flags, unsigned *indent,
+complain_indent (location const *loc, warnings flags, int *indent,
                  const char *message, ...)
 {
   va_list args;
-  indent_ptr = indent;
   va_start (args, message);
-  complains (loc, flags, message, args);
+  complains (loc, indent, flags, message, args);
   va_end (args);
 }
 
 void
-complain_args (location const *loc, warnings w, unsigned *indent,
+complain_args (location const *loc, warnings w, int *indent,
                int argc, char *argv[])
 {
   switch (argc)
@@ -528,24 +546,23 @@ bison_directive (location const *loc, char const *directive)
 void
 deprecated_directive (location const *loc, char const *old, char const *upd)
 {
-  if (feature_flag & feature_caret)
-    complain (loc, Wdeprecated,
-              _("deprecated directive, use %s"),
-              quote_n (1, upd));
-  else
-    complain (loc, Wdeprecated,
-              _("deprecated directive: %s, use %s"),
-              quote (old), quote_n (1, upd));
-  /* Register updates only if -Wdeprecated is enabled.  */
   if (warning_is_enabled (Wdeprecated))
-    fixits_register (loc, upd);
+    {
+      complain (loc, Wdeprecated,
+                _("deprecated directive: %s, use %s"),
+                quote (old), quote_n (1, upd));
+      if (feature_flag & feature_caret)
+        location_caret_suggestion (*loc, upd, stderr);
+      /* Register updates only if -Wdeprecated is enabled.  */
+      fixits_register (loc, upd);
+    }
 }
 
 void
 duplicate_directive (char const *directive,
                      location first, location second)
 {
-  unsigned i = 0;
+  int i = 0;
   if (feature_flag & feature_caret)
     complain_indent (&second, Wother, &i, _("duplicate directive"));
   else
@@ -559,7 +576,7 @@ void
 duplicate_rule_directive (char const *directive,
                           location first, location second)
 {
-  unsigned i = 0;
+  int i = 0;
   complain_indent (&second, complaint, &i,
                    _("only one %s allowed per rule"), directive);
   i += SUB_INDENT;
