@@ -39,21 +39,21 @@
 #include "reduce.h"
 #include "symtab.h"
 
-/* Set of all nonterminals whose language is not empty.  */
+/* Set of nonterminals whose language is not empty.  */
 static bitset N;
 
-/* Set of all rules which have no useless nonterminals in their RHS.  */
+/* Set of rules that have no useless nonterminals in their RHS.  */
 static bitset P;
 
-/* Set of all accessible symbols.  */
+/* Set of accessible symbols.  */
 static bitset V;
 
 /* Set of symbols used to define rule precedence (so they are
    'useless', but no warning should be issued).  */
 static bitset V1;
 
-unsigned nuseless_productions;
-unsigned nuseless_nonterminals;
+int nuseless_productions;
+int nuseless_nonterminals;
 
 #define bitset_swap(Lhs, Rhs)                   \
   do {                                          \
@@ -81,9 +81,11 @@ useful_production (rule_number r, bitset N0)
 }
 
 
-/*---------------------------------------------------------.
-| Remember that rules are 1-origin, symbols are 0-origin.  |
-`---------------------------------------------------------*/
+/*-----------------------------------------------------------------.
+| Compute N, the set of nonterminals whose language is not empty.  |
+|                                                                  |
+| Remember that rules are 1-origin, symbols are 0-origin.          |
+`-----------------------------------------------------------------*/
 
 static void
 useless_nonterminals (void)
@@ -193,10 +195,10 @@ inaccessable_symbols (void)
   bitset_free (P);
   P = Pp;
 
-  unsigned nuseful_productions = bitset_count (P);
+  int nuseful_productions = bitset_count (P);
   nuseless_productions = nrules - nuseful_productions;
 
-  unsigned nuseful_nonterminals = 0;
+  int nuseful_nonterminals = 0;
   for (symbol_number i = ntokens; i < nsyms; ++i)
     nuseful_nonterminals += bitset_test (V, i);
   nuseless_nonterminals = nvars - nuseful_nonterminals;
@@ -248,10 +250,7 @@ reduce_grammar_tables (void)
 
   /* Adjust NRITEMS.  */
   for (rule_number r = nrules; r < nrules + nuseless_productions; ++r)
-    {
-      int length = rule_rhs_length (&rules[r]);
-      nritems -= length + 1;
-  }
+    nritems -= rule_rhs_length (&rules[r]) + 1;
 }
 
 
@@ -259,22 +258,23 @@ reduce_grammar_tables (void)
 | Remove useless nonterminals.  |
 `------------------------------*/
 
+symbol_number *nterm_map = NULL;
+
 static void
 nonterminals_reduce (void)
 {
+  nterm_map = xnmalloc (nvars, sizeof *nterm_map);
   /* Map the nonterminals to their new index: useful first, useless
      afterwards.  Kept for later report.  */
-
-  symbol_number *nontermmap = xnmalloc (nvars, sizeof *nontermmap);
   {
     symbol_number n = ntokens;
     for (symbol_number i = ntokens; i < nsyms; ++i)
       if (bitset_test (V, i))
-        nontermmap[i - ntokens] = n++;
+        nterm_map[i - ntokens] = n++;
     for (symbol_number i = ntokens; i < nsyms; ++i)
       if (!bitset_test (V, i))
         {
-          nontermmap[i - ntokens] = n++;
+          nterm_map[i - ntokens] = n++;
           if (symbols[i]->content->status != used)
             complain (&symbols[i]->location, Wother,
                       _("nonterminal useless in grammar: %s"),
@@ -282,32 +282,30 @@ nonterminals_reduce (void)
         }
   }
 
-
   /* Shuffle elements of tables indexed by symbol number.  */
   {
     symbol **symbols_sorted = xnmalloc (nvars, sizeof *symbols_sorted);
     for (symbol_number i = ntokens; i < nsyms; ++i)
-      symbols[i]->content->number = nontermmap[i - ntokens];
+      symbols[i]->content->number = nterm_map[i - ntokens];
     for (symbol_number i = ntokens; i < nsyms; ++i)
-      symbols_sorted[nontermmap[i - ntokens] - ntokens] = symbols[i];
+      symbols_sorted[nterm_map[i - ntokens] - ntokens] = symbols[i];
     for (symbol_number i = ntokens; i < nsyms; ++i)
       symbols[i] = symbols_sorted[i - ntokens];
     free (symbols_sorted);
   }
 
+  /* Update nonterminal numbers in the RHS of the rules.  LHS are
+     pointers to the symbol structure, they don't need renumbering. */
   {
     for (rule_number r = 0; r < nrules; ++r)
       for (item_number *rhsp = rules[r].rhs; 0 <= *rhsp; ++rhsp)
         if (ISVAR (*rhsp))
-          *rhsp =  symbol_number_as_item_number (nontermmap[*rhsp
-                                                            - ntokens]);
-    accept->content->number = nontermmap[accept->content->number - ntokens];
+          *rhsp = symbol_number_as_item_number (nterm_map[*rhsp - ntokens]);
+    accept->content->number = nterm_map[accept->content->number - ntokens];
   }
 
   nsyms -= nuseless_nonterminals;
   nvars -= nuseless_nonterminals;
-
-  free (nontermmap);
 }
 
 
@@ -322,7 +320,7 @@ reduce_output (FILE *out)
     {
       fprintf (out, "%s\n\n", _("Nonterminals useless in grammar"));
       for (int i = 0; i < nuseless_nonterminals; ++i)
-        fprintf (out, "   %s\n", symbols[nsyms + i]->tag);
+        fprintf (out, "    %s\n", symbols[nsyms + i]->tag);
       fputs ("\n\n", out);
     }
 
@@ -334,7 +332,7 @@ reduce_output (FILE *out)
           if (!b)
             fprintf (out, "%s\n\n", _("Terminals unused in grammar"));
           b = true;
-          fprintf (out, "   %s\n", symbols[i]->tag);
+          fprintf (out, "    %s\n", symbols[i]->tag);
         }
     if (b)
       fputs ("\n\n", out);
@@ -379,23 +377,23 @@ reduce_grammar (void)
   inaccessable_symbols ();
 
   /* Did we reduce something? */
-  if (!nuseless_nonterminals && !nuseless_productions)
-    return;
+  if (nuseless_nonterminals || nuseless_productions)
+    {
+      reduce_print ();
 
-  reduce_print ();
+      if (!bitset_test (N, accept->content->number - ntokens))
+        complain (&startsymbol_loc, fatal,
+                  _("start symbol %s does not derive any sentence"),
+                  startsymbol->tag);
 
-  if (!bitset_test (N, accept->content->number - ntokens))
-    complain (&startsymbol_location, fatal,
-              _("start symbol %s does not derive any sentence"),
-              startsymbol->tag);
-
-  /* First reduce the nonterminals, as they renumber themselves in the
-     whole grammar.  If you change the order, nonterms would be
-     renumbered only in the reduced grammar.  */
-  if (nuseless_nonterminals)
-    nonterminals_reduce ();
-  if (nuseless_productions)
-    reduce_grammar_tables ();
+      /* First reduce the nonterminals, as they renumber themselves in the
+         whole grammar.  If you change the order, nonterms would be
+         renumbered only in the reduced grammar.  */
+      if (nuseless_nonterminals)
+        nonterminals_reduce ();
+      if (nuseless_productions)
+        reduce_grammar_tables ();
+    }
 
   if (trace_flag & trace_grammar)
     {
@@ -433,4 +431,6 @@ reduce_free (void)
   bitset_free (V);
   bitset_free (V1);
   bitset_free (P);
+  free (nterm_map);
+  nterm_map = NULL;
 }
