@@ -24,18 +24,27 @@
 #include "system.h"
 
 #include <assure.h>
+#include <fstrcmp.h>
 #include <hash.h>
 
 #include "complain.h"
+#include "getargs.h"
 #include "gram.h"
+#include "intprops.h"
+#include "quote.h"
 
-/*-------------------------------------------------------------------.
-| Symbols sorted by tag.  Allocated by the first invocation of       |
-| symbols_do, after which no more symbols should be created.         |
-`-------------------------------------------------------------------*/
+
+static struct hash_table *symbol_table = NULL;
+static struct hash_table *semantic_type_table = NULL;
+
+/*----------------------------------------------------------------.
+| Symbols sorted by tag.  Allocated by table_sort, after which no |
+| more symbols should be created.                                 |
+`----------------------------------------------------------------*/
 
 static symbol **symbols_sorted = NULL;
-static symbol **semantic_types_sorted = NULL;
+static semantic_type **semantic_types_sorted = NULL;
+
 
 /*------------------------.
 | Distinguished symbols.  |
@@ -149,8 +158,8 @@ symbol_free (void *ptr)
    declaration first.
 */
 
-static
-void symbols_sort (symbol **first, symbol **second)
+static void
+symbols_sort (symbol **first, symbol **second)
 {
   if (0 < location_cmp ((*first)->location, (*second)->location))
     {
@@ -162,8 +171,8 @@ void symbols_sort (symbol **first, symbol **second)
 
 /* Likewise, for locations.  */
 
-static
-void locations_sort (location *first, location *second)
+static void
+locations_sort (location *first, location *second)
 {
   if (0 < location_cmp (*first, *second))
     {
@@ -279,7 +288,7 @@ static void
 complain_symbol_redeclared (symbol *s, const char *what, location first,
                             location second)
 {
-  unsigned i = 0;
+  int i = 0;
   locations_sort (&first, &second);
   complain_indent (&second, complaint, &i,
                    _("%s redeclaration for %s"), what, s->tag);
@@ -292,7 +301,7 @@ static void
 complain_semantic_type_redeclared (semantic_type *s, const char *what, location first,
                                    location second)
 {
-  unsigned i = 0;
+  int i = 0;
   locations_sort (&first, &second);
   complain_indent (&second, complaint, &i,
                    _("%s redeclaration for <%s>"), what, s->tag);
@@ -304,7 +313,7 @@ complain_semantic_type_redeclared (semantic_type *s, const char *what, location
 static void
 complain_class_redeclared (symbol *sym, symbol_class class, location second)
 {
-  unsigned i = 0;
+  int i = 0;
   complain_indent (&second, complaint, &i,
                    class == token_sym
                    ? _("symbol %s redeclared as a token")
@@ -317,6 +326,55 @@ complain_class_redeclared (symbol *sym, symbol_class class, location second)
     }
 }
 
+static const symbol *
+symbol_from_uniqstr_fuzzy (const uniqstr key)
+{
+  aver (symbols_sorted);
+#define FSTRCMP_THRESHOLD 0.6
+  double best_similarity = FSTRCMP_THRESHOLD;
+  const symbol *res = NULL;
+  size_t count = hash_get_n_entries (symbol_table);
+  for (int i = 0; i < count; ++i)
+    {
+      symbol *sym = symbols_sorted[i];
+      if (STRNEQ (key, sym->tag)
+          && (sym->content->status == declared
+              || sym->content->status == undeclared))
+        {
+          double similarity = fstrcmp_bounded (key, sym->tag, best_similarity);
+          if (best_similarity < similarity)
+            {
+              res = sym;
+              best_similarity = similarity;
+            }
+        }
+    }
+  return res;
+}
+
+static void
+complain_symbol_undeclared (symbol *sym)
+{
+  assert (sym->content->status != declared);
+  const symbol *best = symbol_from_uniqstr_fuzzy (sym->tag);
+  if (best)
+    {
+      complain (&sym->location,
+                sym->content->status == needed ? complaint : Wother,
+                _("symbol %s is used, but is not defined as a token"
+                  " and has no rules; did you mean %s?"),
+                quote_n (0, sym->tag),
+                quote_n (1, best->tag));
+      if (feature_flag & feature_caret)
+        location_caret_suggestion (sym->location, best->tag, stderr);
+    }
+  else
+    complain (&sym->location,
+              sym->content->status == needed ? complaint : Wother,
+              _("symbol %s is used, but is not defined as a token"
+                " and has no rules"),
+              quote (sym->tag));
+}
 
 void
 symbol_location_as_lhs_set (symbol *sym, location loc)
@@ -463,8 +521,9 @@ symbol_class_set (symbol *sym, symbol_class class, location loc, bool declaring)
         {
           if (s->status == declared)
             {
-              unsigned i = 0;
-              complain (&loc, Wother, _("symbol %s redeclared"), sym->tag);
+              int i = 0;
+              complain_indent (&loc, Wother, &i,
+                               _("symbol %s redeclared"), sym->tag);
               i += SUB_INDENT;
               complain_indent (&sym->location, Wother, &i,
                                _("previous declaration"));
@@ -491,6 +550,9 @@ symbol_user_token_number_set (symbol *sym, int user_token_number, location loc)
            && *user_token_numberp != user_token_number)
     complain (&loc, complaint, _("redefining user token number of %s"),
               sym->tag);
+  else if (user_token_number == INT_MAX)
+    complain (&loc, complaint, _("user token number of %s too large"),
+              sym->tag);
   else
     {
       *user_token_numberp = user_token_number;
@@ -513,18 +575,13 @@ symbol_user_token_number_set (symbol *sym, int user_token_number, location loc)
 | nonterminal.                                              |
 `----------------------------------------------------------*/
 
-static inline bool
+static void
 symbol_check_defined (symbol *sym)
 {
   sym_content *s = sym->content;
   if (s->class == unknown_sym)
     {
-      assert (s->status != declared);
-      complain (&sym->location,
-                s->status == needed ? complaint : Wother,
-                _("symbol %s is used, but is not defined as a token"
-                  " and has no rules"),
-                  sym->tag);
+      complain_symbol_undeclared (sym);
       s->class = nterm_sym;
       s->number = nvars++;
     }
@@ -540,11 +597,9 @@ symbol_check_defined (symbol *sym)
       if (sem_type)
         sem_type->status = declared;
     }
-
-  return true;
 }
 
-static inline bool
+static void
 semantic_type_check_defined (semantic_type *sem_type)
 {
   /* <*> and <> do not have to be "declared".  */
@@ -563,24 +618,8 @@ semantic_type_check_defined (semantic_type *sem_type)
     complain (&sem_type->location, Wother,
               _("type <%s> is used, but is not associated to any symbol"),
               sem_type->tag);
-
-  return true;
 }
 
-static bool
-symbol_check_defined_processor (void *sym, void *null ATTRIBUTE_UNUSED)
-{
-  return symbol_check_defined (sym);
-}
-
-static bool
-semantic_type_check_defined_processor (void *sem_type,
-                                       void *null ATTRIBUTE_UNUSED)
-{
-  return semantic_type_check_defined (sem_type);
-}
-
-
 /*-------------------------------------------------------------------.
 | Merge the properties (precedence, associativity, etc.) of SYM, and |
 | its string-named alias STR; check consistency.                     |
@@ -647,7 +686,7 @@ symbol_make_alias (symbol *sym, symbol *str, location loc)
 | into FDEFINES.  Put in SYMBOLS.                                    |
 `-------------------------------------------------------------------*/
 
-static inline bool
+static void
 symbol_pack (symbol *this)
 {
   aver (this->content->number != NUMBER_UNDEFINED);
@@ -655,19 +694,12 @@ symbol_pack (symbol *this)
     this->content->number += ntokens;
 
   symbols[this->content->number] = this->content->symbol;
-  return true;
-}
-
-static bool
-symbol_pack_processor (void *this, void *null ATTRIBUTE_UNUSED)
-{
-  return symbol_pack (this);
 }
 
 static void
 complain_user_token_number_redeclared (int num, symbol *first, symbol *second)
 {
-  unsigned i = 0;
+  int i = 0;
   symbols_sort (&first, &second);
   complain_indent (&second->location, complaint, &i,
                    _("user token number %d redeclaration for %s"),
@@ -682,7 +714,7 @@ complain_user_token_number_redeclared (int num, symbol *first, symbol *second)
 | Put THIS in TOKEN_TRANSLATIONS if it is a token.  |
 `--------------------------------------------------*/
 
-static inline bool
+static void
 symbol_translation (symbol *this)
 {
   /* Nonterminal? */
@@ -699,14 +731,6 @@ symbol_translation (symbol *this)
         token_translations[this->content->user_token_number]
           = this->content->number;
     }
-
-  return true;
-}
-
-static bool
-symbol_translation_processor (void *this, void *null ATTRIBUTE_UNUSED)
-{
-  return symbol_translation (this);
 }
 
 
@@ -717,9 +741,6 @@ symbol_translation_processor (void *this, void *null ATTRIBUTE_UNUSED)
 /* Initial capacity of symbol and semantic type hash table.  */
 #define HT_INITIAL_CAPACITY 257
 
-static struct hash_table *symbol_table = NULL;
-static struct hash_table *semantic_type_table = NULL;
-
 static inline bool
 hash_compare_symbol (const symbol *m1, const symbol *m2)
 {
@@ -784,6 +805,23 @@ symbols_new (void)
                                    hash_symbol_hasher,
                                    hash_symbol_comparator,
                                    symbol_free);
+
+  /* Construct the accept symbol. */
+  accept = symbol_get ("$accept", empty_loc);
+  accept->content->class = nterm_sym;
+  accept->content->number = nvars++;
+
+  /* Construct the error token */
+  errtoken = symbol_get ("error", empty_loc);
+  errtoken->content->class = token_sym;
+  errtoken->content->number = ntokens++;
+
+  /* Construct a token that represents all undefined literal tokens.
+     It is always token number 2.  */
+  undeftoken = symbol_get ("$undefined", empty_loc);
+  undeftoken->content->class = token_sym;
+  undeftoken->content->number = ntokens++;
+
   semantic_type_table = hash_xinitialize (HT_INITIAL_CAPACITY,
                                           NULL,
                                           hash_semantic_type_hasher,
@@ -905,36 +943,25 @@ symbols_free (void)
 }
 
 
-/*---------------------------------------------------------------.
-| Look for undefined symbols, report an error, and consider them |
-| terminals.                                                     |
-`---------------------------------------------------------------*/
-
 static int
-symbols_cmp (symbol const *a, symbol const *b)
+symbol_cmp (void const *a, void const *b)
 {
-  return strcmp (a->tag, b->tag);
+  return location_cmp ((*(symbol * const *)a)->location,
+                       (*(symbol * const *)b)->location);
 }
 
-static int
-symbols_cmp_qsort (void const *a, void const *b)
-{
-  return symbols_cmp (*(symbol * const *)a, *(symbol * const *)b);
-}
+/* Store in *SORTED an array of pointers to the symbols contained in
+   TABLE, sorted (alphabetically) by tag. */
 
 static void
-symbols_do (Hash_processor processor, void *processor_data,
-            struct hash_table *table, symbol ***sorted)
+table_sort (struct hash_table *table, symbol ***sorted)
 {
+  aver (!*sorted);
   size_t count = hash_get_n_entries (table);
-  if (!*sorted)
-    {
-      *sorted = xnmalloc (count, sizeof **sorted);
-      hash_get_entries (table, (void**)*sorted, count);
-      qsort (*sorted, count, sizeof **sorted, symbols_cmp_qsort);
-    }
-  for (size_t i = 0; i < count; ++i)
-    processor ((*sorted)[i], processor_data);
+  *sorted = xnmalloc (count + 1, sizeof **sorted);
+  hash_get_entries (table, (void**)*sorted, count);
+  qsort (*sorted, count, sizeof **sorted, symbol_cmp);
+  (*sorted)[count] = NULL;
 }
 
 /*--------------------------------------------------------------.
@@ -945,10 +972,20 @@ symbols_do (Hash_processor processor, void *processor_data,
 void
 symbols_check_defined (void)
 {
-  symbols_do (symbol_check_defined_processor, NULL,
-              symbol_table, &symbols_sorted);
-  symbols_do (semantic_type_check_defined_processor, NULL,
-              semantic_type_table, &semantic_types_sorted);
+  table_sort (symbol_table, &symbols_sorted);
+  /* semantic_type, like symbol, starts with a 'tag' field and then a
+     'location' field.  And here we only deal with arrays/hashes of
+     pointers, sizeof is not an issue.
+
+     So instead of implementing table_sort (and symbol_cmp) once for
+     each type, let's lie a bit to the typing system, and treat
+     'semantic_type' as if it were 'symbol'. */
+  table_sort (semantic_type_table, (symbol ***) &semantic_types_sorted);
+
+  for (int i = 0; symbols_sorted[i]; ++i)
+    symbol_check_defined (symbols_sorted[i]);
+  for (int i = 0; semantic_types_sorted[i]; ++i)
+    semantic_type_check_defined (semantic_types_sorted[i]);
 }
 
 /*------------------------------------------------------------------.
@@ -989,7 +1026,13 @@ symbols_token_translations_init (void)
     {
       sym_content *this = symbols[i]->content;
       if (this->user_token_number == USER_NUMBER_UNDEFINED)
-        this->user_token_number = ++max_user_token_number;
+        {
+          IGNORE_TYPE_LIMITS_BEGIN
+          if (INT_ADD_WRAPV (max_user_token_number, 1, &max_user_token_number))
+            complain (NULL, fatal, _("token number too large"));
+          IGNORE_TYPE_LIMITS_END
+          this->user_token_number = max_user_token_number;
+        }
       if (this->user_token_number > max_user_token_number)
         max_user_token_number = this->user_token_number;
     }
@@ -999,10 +1042,10 @@ symbols_token_translations_init (void)
 
   /* Initialize all entries for literal tokens to the internal token
      number for $undefined, which represents all invalid inputs.  */
-  for (int i = 0; i < max_user_token_number + 1; i++)
+  for (int i = 0; i < max_user_token_number + 1; ++i)
     token_translations[i] = undeftoken->content->number;
-  symbols_do (symbol_translation_processor, NULL,
-              symbol_table, &symbols_sorted);
+  for (int i = 0; symbols_sorted[i]; ++i)
+    symbol_translation (symbols_sorted[i]);
 }
 
 
@@ -1015,7 +1058,8 @@ void
 symbols_pack (void)
 {
   symbols = xcalloc (nsyms, sizeof *symbols);
-  symbols_do (symbol_pack_processor, NULL, symbol_table, &symbols_sorted);
+  for (int i = 0; symbols_sorted[i]; ++i)
+    symbol_pack (symbols_sorted[i]);
 
   /* Aliases leave empty slots in symbols, so remove them.  */
   {
